#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dos.h>
#include "bzboot.h"

#define DEFAULT_IMAGE "bzImage"
#define SECT_SIZE 512
#define ADDR_TYPE_LOADER 0x10
#define ADDR_MAJOR 509
#define ADDR_MINOR 508
#define CMDLINE_SIZE 256
#define CMDLINE_ADDR 0x99000
#define CMDLINE_PTR 0x28

# define out_b(addr, data) outportb(addr, data)
# define out_w(addr, data) outport(addr, data)
# define DI disable()
# define EI enable()
# define bios(no, iregp, oregp) int86(no, iregp, oregp)
typedef union REGS regset;


unsigned char *bootsect, *setup, *cmdline;
unsigned short setup_size;
unsigned short setup_buf_size;
unsigned long xms_image_addr = 0;
unsigned long kernel_size;

static unsigned short int cregs0[]={
	0x0040,0x0320,0xffff,0xffff,0x035f,0x0000,0x0010,0x0000,
	0x0370,0x009c,0x031c,0x009c,0x031c,0x0040,0x0360,0x0040,
	0x0360,0x0000,0x009c,0x0000,0x0050,0x0000,0x009c,0x0000,
	0x0080,0x004a,0x0001,0x0000,0x801f,0x0003,0x0000,0x0150
};
static unsigned short int cregs9[]={
	0x0060,0x02c0,0xffff,0xffff,0x031f,0x0000,0x0004,0x0000,
	0x0419,0x008a,0x030a,0x008a,0x030a,0x0046,0x0406,0x0046,
	0x0406,0x0000,0x008a,0x0000,0x0080,0x0000,0x008a,0x0000,
	0x0080,0x0058,0x0001,0x0000,0x800f,0x0002,0x0000,0x0192

};
static unsigned char sregs0[]={
	0x15,0x09
};
static unsigned char sregs9[]={
	0x15,0x08
};

void load_high(fname)
     char *fname;
{
    struct stat st;

    /* get filesize */
    if (stat(fname, &st) < 0){
  	perror(fname);
  	return;
    }

    kernel_size = st.st_size;
    printf("kernel size = %ld b\n", kernel_size);

    xmsload(&xms_image_addr, kernel_size);
    if (!xms_image_addr) {
	cocoload(&xms_image_addr, kernel_size);
	if (!xms_image_addr) {
	    (void) fprintf(stderr, "can't load with xms.\n");
	    exit (-1);
	}
    }
}

void read_setup()
{
    int setup_sects;
    unsigned char *p, *q;
    int i;

    setup_sects = bootsect[497];
    setup_buf_size = (setup_sects+1) * SECT_SIZE;
    setup_size = setup_sects * SECT_SIZE;

    (void) printf("setup_sects = %d setup_size=%d\n", setup_sects, setup_size);

    setup = (unsigned char *) malloc(setup_buf_size+1);
    p = setup + SECT_SIZE;
    (void) read(fd, p, setup_size);

    /* type_of_loader */
    p[ADDR_TYPE_LOADER] = 0x20;

    /* transfer bootsect to setup */
    p = setup;
    q = bootsect;
    for(i=0; i<SECT_SIZE; i++){
	*p++ = *q++;
    }
}

void make_cmdline(n,argc,argv)
     int n, argc;
     char **argv;
{
    int i,j;
    long cmdline_addr;
    unsigned char *p, *q;
    int init = 0;

    if (n<argc-1) {
	p = setup + SECT_SIZE + CMDLINE_PTR;
	cmdline_addr = CMDLINE_ADDR;
	for(i=0; i<4; i++){
	    j = cmdline_addr & 0xff;
	    *p++ = j;
	    cmdline_addr >>=8;
	}

	p = cmdline = (unsigned char *) malloc(CMDLINE_SIZE+1);
	for(i=n+1; i<argc; i++){
	    if (!init) {
		init = 1;
	    } else {
		*p++ = ' ';
	    }
	    q = argv[i];
	    while(*q){
		*p++ = *q++;
	    }
	    *p = 0;
	}
    }
}

void read_bootsect(m,n)
     int m,n;
{
    bootsect = (unsigned char *) malloc(SECT_SIZE);
    (void) read(fd, bootsect, SECT_SIZE);

    if (m || n) {
	bootsect[ADDR_MAJOR] = m;
	bootsect[ADDR_MINOR] = n;
    }
}

void open_file(fname)
     char *fname;
{
    if (!(fd = open(fname, O_RDONLY|O_BINARY))) {
	(void) fprintf(stderr,"can't open [%s]\n", fname);
	exit(-1);
    }
}

void clear_screen()
{
    regset ireg,oreg;

    ireg.x.ax=0x0b01;	/* cursor off */
    bios(0x91,&ireg,&oreg);
    ireg.x.ax=0x1b00;	/* set console control flag */
    bios(0x91,&ireg,&oreg);
    ireg.x.ax=0x0200;	/* display all clear */
    bios(0x91,&ireg,&oreg);
    ireg.x.cx=0x4000;    /* wait CX reg x 10 us */
    bios(0xfd,&ireg,&oreg);
}

void setup_crtc(int sw)
{
	int i;
	unsigned short int *cregs;
	unsigned char *sregs;

	if(sw==0) {
		cregs=cregs0;
		sregs=sregs0;
		out_b(0xff82,0x47);
		out_b(0xff81,0x0f);
	}
	else {
/*	640X480(1024X512)X(16/4096)colorsX2pages */
		cregs=cregs9;
		sregs=sregs9;
		out_b(0xff82,0x67);
		out_b(0xff81,0x0f);
		out_b(0xfda0,0x0f);
		out_b(0x458,0x0);
		out_b(0x45a,0xff);
		out_b(0x45b,0xff);
		out_b(0x458,0x1);
		out_b(0x45a,0xff);
		out_b(0x45b,0xff);
	}

	DI;
	for(i=0;i<0x20;i++) {
		if(cregs[i]==0xffff) continue;
		out_b(0x440,i);
		out_w(0x442,cregs[i]);
	}
	for(i=0;i<2;i++) {
		out_b(0x448,i);
		out_b(0x44a,sregs[i]);
	}
	EI;
}

int main(argc,argv)
     int argc;
     char **argv;
{
    int major = 0;
    int minor = 0;
    int i;
    char *fname;

    i = 1;
    while (i<=argc && *argv[i] == '-') {
	switch (argv[i][1]) {
	 case 'm':
	     i++;
	     sscanf(argv[i], "%d", &major);
	     break;
	 case 'n':
	     i++;
	     sscanf(argv[i], "%d", &minor);
	     break;
	 default:
	     (void) fprintf(stderr, "unknown option [%s]\n", argv[i]);
	     exit (-1);
	}
	i++;
    }
    if (i > argc || argc == 1) {
	fname = DEFAULT_IMAGE;
    } else {
	fname = argv[i];
    }
    if (!fname) {
	fname = DEFAULT_IMAGE;
    }

    printf("major=%d minor=%d image=%s\n", major, minor, fname);

    open_file(fname);
    read_bootsect(major, minor);
    read_setup();

    make_cmdline(i,argc,argv);

    load_high(fname);

    printf("push return");
    getchar();

    clear_screen();
    setup_crtc(1);

    terminate_ems();
    exec_setup(setup, setup_buf_size,
	       xms_image_addr, kernel_size, cmdline);

    return 0;
}
