#define EXTERN
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include "bzboot.h"

#define BUFSIZE (1024*8)
#define DEFAULT_MIN_ADDR 0x280000
#define PAGESIZE 4096
#define PAGEMASK (~((unsigned long)(PAGESIZE-1)))

static char *buf=NULL;
static unsigned long phys = 0L;
#ifdef XMS
static int xmshndl = -1;

static struct xms_move {
  unsigned long size;
  short src_hndl;
  unsigned long src_offset;
  short dst_hndl;
  unsigned long dst_offset;
} move_param = {
  (unsigned long)BUFSIZE, 0
};
#endif /* XMS */

static void ext_terminate(void)
{
#ifdef XMS
  if (xmshndl != -1){
    xms_unlock(xmshndl);
    xms_free(xmshndl);
  }
#endif /* XMS */
#ifdef COCO
  if (phys != 0L){
    coco_free(phys);
  }
#endif /* COCO */
  if(buf!=NULL) {
    free(buf);
  }
}

static void on_int(int sig)
{
  ext_terminate();
  exit(2);
}

void
#ifdef COCO
cocoload
#endif /* COCO */
#ifdef XMS
xmsload
#endif /* XMS */
(unsigned long *addr,unsigned long size)
{
  unsigned long n;
  unsigned long load_addr, min_addr, p;
#ifdef XMS
  unsigned long maxfree, offset;
#endif /* XMS */
#ifdef COCO
  unsigned short maxfree;
  unsigned long src;
#endif /* COCO */
#ifdef XMS
  extern int xms_check(void), xms_alloc(unsigned long);
  extern int xms_copy(struct xms_move *);
  extern unsigned long xms_maxfree(void), xms_lock(int);
#endif /* XMS */
#ifdef COCO
  extern unsigned short coco_maxfree(void);
  extern unsigned long coco_alloc(unsigned short);
  extern int coco_free(unsigned long);
  extern int coco_copy(unsigned long, unsigned long, unsigned short);
#endif /* COCO */

  if((buf=malloc(BUFSIZE))==NULL) {
      (void) fprintf(stderr, "can't allocate memory.\n");
      return;
  }

#ifdef XMS
  move_param.src_offset=(unsigned long)(char far *)buf;
#endif

  min_addr = DEFAULT_MIN_ADDR;

  load_addr = *addr;
  
  signal(SIGINT, on_int);
  
  /* get memory */
#ifdef XMS
  if (xms_check() < 0){
    fprintf(stderr, "XMS not found.\n");
    return;
  }
  if ((maxfree = xms_maxfree()) == 0xffffffff){
  	fprintf(stderr, "xms_maxfree: error\n");
  	return;
  }
  fprintf(stderr, "largest free memory block = %ld KB\n", maxfree);
  if ((xmshndl = xms_alloc(maxfree)) == -1){
  	fprintf(stderr, "xms_alloc: error\n");
  	return;
  }
  
  /* get physical load_address */
  if ((phys = xms_lock(xmshndl)) == 0xffffffff){
    fprintf(stderr, "xms_lock: error\n");
  	xms_free(xmshndl);
    return;
  }
#endif /* XMS */
#ifdef COCO
  maxfree = coco_maxfree();
  fprintf(stderr, "largest free memory block = %u KB\n", maxfree);
  if ((phys = coco_alloc(maxfree)) == 0L){
  	fprintf(stderr, "coco_alloc: error\n");
  	return;
  }
#endif /* COCO */
  fprintf(stderr, "    physical load_address = 0x%lx\n", phys);
  (void) fprintf(stderr,"max addr=%lx\n", maxfree*1024L+phys);
  if (load_addr == 0L){
    load_addr = (phys + maxfree * 1024L - (size + BUFSIZE))&PAGEMASK;
    if (load_addr < phys  ||  load_addr < min_addr){
      fprintf(stderr, "Not enough memory for bzImage.\n");
      ext_terminate();
      return;
    }
  } else {
    load_addr&=PAGEMASK;
    if (load_addr < phys  ||  phys + maxfree*1024L < load_addr + size){
      fprintf(stderr, "load_addr = 0x%lx, must be in 0x%lx --- 0x%lx\n",
              load_addr, phys, phys + maxfree*1024L-size);
      ext_terminate();
      return;
    }
  }
  fprintf(stderr, "Load bzImage at 0x%lx.\n", load_addr);
#ifdef XMS
  offset = load_addr - phys;
#endif /* XMS */

  /* read bzImage to xms memory */
#ifdef XMS
  move_param.dst_hndl = xmshndl;
  p = offset;
#endif /* XMS */
#ifdef COCO
  p = (unsigned long)(char far *)buf;
  src = ((p >> 12) & ~0xfL) + (p & 0xffffL);  /* make linear load_address */
  p = load_addr;
#endif /* COCO */  
  while ((n = (unsigned long) read(fd, buf, BUFSIZE)) > 0) {
#ifdef XMS
      move_param.dst_offset = p;
      if (xms_copy(&move_param) != 1){
	  fprintf(stderr, "xms_copy: error\n");
	  ext_terminate();
	  return;
      }
#endif /* XMS */
#ifdef COCO
  	coco_copy(src, p, BUFSIZE);
#endif /* COCO */
	(void) fprintf(stderr,".");
  	p += (unsigned long) n;
  }
  (void) fprintf(stderr,"\n");

#ifdef XMS
  p -= offset;
#endif /* XMS */
#ifdef COCO
  p -= load_addr;
#endif /* COCO */
  fprintf(stderr, "%ld bytes read.\n", p);
  /*
  if (p != st.st_size){
    fprintf(stderr, "read error: expect=%ld byte, actual=%ld byte\n",
            st.st_size, p);
    ext_terminate();
    return;
  }
  */

  *addr=load_addr;
  free(buf);

  return;
}

